/* menu.c
   menus for luminesweeper

Copyright 2005 Damian Yerrick

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Author contact information is in readme.txt

*/

#include "util.h"

GameSpecificGlobals game_globals;

const GBFS_FILE *data_gbfs;
const GBFS_FILE *Lskins_gbfs;

extern const char _end[];

static void setup_isr(void)
{
  /* set the ISR */
  SET_MASTER_ISR(isr);
  /* turn on interrupt sources */
  LCDSTAT = LCDSTAT_VBLIRQ;  /* vbl interrupt */
  /* turn on interrupt controller */
  INTMASK = INT_VBLANK;
  INTENABLE = 1;
}


/* there are three slots for todbgs.

*/
void load_todbg(unsigned int slot, const unsigned char *name)
{
  /* decompression buffer at EWRAM Top - 22 KB */
  u32 *decomp_buf = (u32 *)(0x02040000 - 22*1024);
  unsigned int tilebase = ((65536 - 19264 * (3 - slot)) >> 5) & 0x01ff;
  unsigned int tileno = tilebase;
  unsigned int namebase = slot;
  const u32 *lz = gbfs_get_obj(data_gbfs, name, NULL);
  unsigned int x, y;

  if(!lz)
    return;

  BGSCROLL[slot].x = 0;
  BGSCROLL[slot].y = 0;
  BGCTRL[slot] = BGCTRL_NAME(namebase) | BGCTRL_PAT(slot) | BGCTRL_16C;

  gblz_unpack(lz, decomp_buf);
  dma_memcpy(PALRAM + 16 * slot, decomp_buf, 32);
  dma_memcpy(PATRAM4(slot, tilebase), decomp_buf+8, 19200);

  for(y = 0; y < 20; y++)
  {
    for(x = 0; x < 30; x++)
      MAP[namebase][y][x] = (tileno++) | (slot << 12);
    MAP[namebase][y][30] = tilebase + 600;
    MAP[namebase][y][31] = tilebase + 600;
  }

  {
    /* set up tiles 600 and 601 */
    u32 *dst = PATRAM4(slot, tileno);
    for(y = 8; y > 0; y--)
      *dst++ = 0x00000000;
    for(y = 8; y > 0; y--)
      *dst++ = 0xcccccccc;

    /* clear the bottom rows of the map */
    dst = (u32 *)MAP[namebase][20];
    tileno |= tileno << 16;
    for(y = 16*12; y > 0; y--)
      *dst++ = tileno;

  }
}


/* menu_set_window() *******************
   Sets window 0 to cover a 240x160 space whose upper left corner
   is at (left, top).
*/
static
void menu_set_window(int left, int top)
{
  int right = left + 240;
  int bottom = top + 160;

  /* clip to GBA size */
  if(left < 0)
    left = 0;
  if(right > 240)
    right = 240;
  if(top < 0)
    top = 0;
  if(bottom > 160)
    bottom = 160;

  WIN_X[0] = WIN_DIM(left & 0xff, right & 0xff);
  WIN_Y[0] = WIN_DIM(top & 0xff, bottom & 0xff);
}

static
void menu_scroll(int xstart, int ystart, int xend, int yend)
{
  int t;

  for(t = 1; t <= 16; t++)
  {
    int x = xstart + (((xend - xstart) * t) >> 4);
    int y = ystart + (((yend - ystart) * t) >> 4);
    int bg1_x = (x > 0) ? x - 240 : (x == 0) ? 0 : x + 240;
    int bg1_y = (y > 0) ? y - 160 : (y == 0) ? 0 : y + 160;

    u.retraces++;
    wait4vbl();

    BGSCROLL[0].x = x;
    BGSCROLL[0].y = y;
    BGSCROLL[1].x = bg1_x;
    BGSCROLL[1].y = bg1_y;
    menu_set_window(-x, -y);
  }
}


/* menu_loading() ***************
   Fake a loading screen as if this were a psp game.
   If the player presses a button in mask, stop and return the
   button pressed (JOY_A, etc).
   Otherwise, return 0.
*/
static
int menu_loading(unsigned int mask)
{
  unsigned int last_joy = 0x3ff;
  unsigned int t;

  for(t = 0; t < 120; t++)
  {
    unsigned int j = JOY ^ 0x3ff;
    unsigned int jnew = j & ~last_joy;
    last_joy = j;

    u.retraces++;
    wait4vbl();
    MAP[1][19][t >> 2] = 1536-603;
    if(jnew & mask)
      return jnew;
  }
  return 0;
}


int main(void)
{
  const unsigned char *next_fs;

  int done = -1;

  LCDMODE = 0;
  PALRAM[0] = RGB(0, 0, 0);
  data_gbfs = find_first_gbfs_file(_end);
  if(!data_gbfs)
  {
    PALRAM[0] = RGB(31, 0, 0);
    setup_isr();
    while(1)
      wait4vbl();
  }

  /* Handle Lskins */
  next_fs = skip_gbfs_file(data_gbfs);
  /* if it doesn't fit within EWRAM then it's probably copied from ROM */
  if(next_fs[0] == 'P'
     && next_fs < ROM
     && (const unsigned char *)skip_gbfs_file((const GBFS_FILE *)next_fs) > EWRAM + 230*1024)
    next_fs += ROM - EWRAM;
  /* if Lskins is not present, crash */
  if(next_fs[0] != 'P')
  {
    PALRAM[0] = RGB(31, 0, 0);
    setup_isr();
    while(1)
      wait4vbl();
  }
  Lskins_gbfs = (const GBFS_FILE *)next_fs;

  next_fs = skip_gbfs_file(Lskins_gbfs);
  if(next_fs < ROM)
    next_fs += ROM - EWRAM;
  if(next_fs[0] == 'P')
    gsm_fs = (const GBFS_FILE *)next_fs;

  srand(0xf0132);
  clear_keybuf();
  init_sound();

  setup_isr();

  while(1)
  {
    if(done < 0)
    {
      LCDMODE = 0;
      load_todbg(0, "title.lz");
      u.retraces++;
      wait4vbl();
      LCDMODE = 0 | LCDMODE_BG0 | LCDMODE_BG1 | LCDMODE_WIN0;
      WININ = WININ_W0BG0 | WININ_OUTBG1;
      menu_set_window(0, 0);
    }
    done = 0;

    do {
      u.retraces++;
      wait4vbl();
      make_repeats(JOY ^ 0x3ff);    
      if(u.repeats[4] == 1)  /* RIGHT */
        done = 4;
      else if(u.repeats[5] == 1)  /* LEFT */
        done = 5;
    } while(!done);

    switch(done)
    {
    case 5:
      load_todbg(1, "Lloading.lz");
      menu_scroll(0, 0, -240, 0);
#if 1
      if(!menu_loading(JOY_RIGHT))
      {
        Lplay();
        done = -1;
      }
      else
        menu_scroll(-240, 0, 0, 0);
#else
      while(JOY & JOY_RIGHT)
      {
        u.retraces++;
        wait4vbl();
      }
      menu_scroll(-240, 0, 0, 0);
#endif
      break;
    case 4:
      load_todbg(1, "Rloading.lz");
      menu_scroll(0, 0, 240, 0);
      if(!menu_loading(JOY_LEFT))
      {
        Rplay();
        done = -1;
      }
      else
        menu_scroll(240, 0, 0, 0);
      break;
    default:
      break;
    }
  }
}

